@@ -1,7 +1,7 @@
 '''test_decoder.py: Contains the set of tests for the decoder class in the Python FIT SDK'''
 
 ###########################################################################################
-# Copyright 2023 Garmin International, Inc.
+# Copyright 2024 Garmin International, Inc.
 # Licensed under the Flexible and Interoperable Data Transfer (FIT) Protocol License; you
 # may not use this file except in compliance with the Flexible and Interoperable Data
 # Transfer (FIT) Protocol License.
@@ -11,7 +11,8 @@
 from datetime import datetime, timezone
 
 import pytest
-from garmin_fit_sdk import Decoder, Stream
+from garmin_fit_sdk import Decoder, Stream, CrcCalculator
+from garmin_fit_sdk.decoder import DecodeMode  
 
 from tests.data import Data
 
@@ -75,6 +76,107 @@ class TestDecoderConstructor:
     except RuntimeError:
         assert True
 
+class TestSkipHeaderDecodeMode:
+    '''Set of tests that test the fuctionality of the skip header decode mode'''
+    def test_invalid_header_with_skip_header(self):
+        '''Tests that file with invalid header should not fail when decode mode is skip header'''
+        stream = Stream.from_byte_array(Data.fit_file_short_invalid_header)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read(decode_mode = DecodeMode.SKIP_HEADER)
+
+        assert len(errors) == 0
+        assert len(messages['file_id_mesgs']) == 1
+
+    def test_invalid_header_without_skip_header(self):
+        '''Tests that file with invalid header should fail when decode mode is normal'''
+        stream = Stream.from_byte_array(Data.fit_file_short_invalid_header)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read()
+
+        assert len(errors) == 1
+
+    def test_valid_header_with_skip_header(self):
+        '''Tests that file with valid header should not fail when decode mode is skip header'''
+        stream = Stream.from_byte_array(Data.fit_file_short)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read(decode_mode = DecodeMode.SKIP_HEADER)
+
+        assert len(errors) == 0
+        assert len(messages['file_id_mesgs']) == 1
+
+    def test_invalid_crc_with_skip_header(self):
+        '''Tests that file with invalid CRC should not fail when decode mode is skip header'''
+        stream = Stream.from_byte_array(Data.fit_file_short_new_invalid_crc)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read(decode_mode = DecodeMode.SKIP_HEADER)
+
+        assert len(errors) == 0
+        assert len(messages['file_id_mesgs']) == 1
+
+class TestDataOnlyDecodeMode:
+    '''Set of tests that test the fuctionality of the data only decode mode'''
+    def test_no_header_with_data_only(self):
+        '''Tests that file with no header should not fail when decode mode is data only'''
+        stream = Stream.from_byte_array(Data.fit_file_short_data_only)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read(decode_mode = DecodeMode.DATA_ONLY)
+
+        assert len(errors) == 0
+        assert len(messages['file_id_mesgs']) == 1
+    
+    def test_no_header_without_data_only(self):
+        '''Tests that file with no header fails when decode mode is data only'''
+        stream = Stream.from_byte_array(Data.fit_file_short_data_only)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read()
+
+        assert len(errors) == 1
+        
+    def test_invalid_crc_with_data_only(self):
+        '''Tests that file with invalid CRC should not fail when decode mode is data only'''
+        stream = Stream.from_byte_array(Data.fit_file_short_new_invalid_crc[14:])
+        decoder = Decoder(stream)
+        messages, errors = decoder.read(decode_mode = DecodeMode.DATA_ONLY)
+
+        assert len(errors) == 0
+        assert len(messages['file_id_mesgs']) == 1
+
+class TestReadFileHeader:
+    '''Set of tests that test the functionality of reading file headers and the File Header class'''
+    def test_read_file_header(self):
+        '''Tests reading the file header with the decoder and decoding the profile and protocol versions.'''
+        stream = Stream.from_byte_array(Data.fit_file_minimum)
+        decoder = Decoder(stream)
+
+        file_header = decoder.read_file_header(stream)
+
+        assert file_header.header_size == 14
+        assert file_header.protocol_version == 32
+        assert file_header.profile_version == 2187
+        assert file_header.data_size == 0
+        assert file_header.data_type == [b'.FIT']
+        assert file_header.header_crc == 18573
+        assert file_header.file_total_size == 14
+
+    def test_read_file_header_dict(self):
+        '''Tests reading the file header and converting the class to a dictionary.'''
+        stream = Stream.from_byte_array(Data.fit_file_minimum)
+        decoder = Decoder(stream)
+
+        file_header = decoder.read_file_header(stream)
+        file_header_dict = file_header.get_dict()
+
+        protocol_version = (file_header.protocol_version >> 4) + ((file_header.protocol_version & 0x0F) / 10)
+        profile_version = file_header.profile_version / 1000 if file_header.profile_version > 2199 else 100
+
+        assert file_header.header_size == file_header_dict['header_size']
+        assert protocol_version == file_header_dict['protocol_version']
+        assert profile_version == file_header_dict['profile_version']
+        assert file_header.data_size == file_header_dict['data_size']
+        assert file_header.data_type == file_header_dict['data_type']
+        assert file_header.header_crc == file_header_dict['header_crc']
+        assert file_header.file_total_size == file_header_dict['file_total_size']
+
 class TestDecoderRead():
     '''Set of tests that verify the validity and accuracy of the decoder when reading files.'''
     @pytest.mark.parametrize(
@@ -94,6 +196,25 @@ class TestDecoderRead():
         assert len(errors) == 0
         assert decoder.get_num_messages() == num_messages
 
+    def test_stream_not_reset(self):
+        '''Tests that the decoder does not reset the stream before decoding.'''
+        stream = Stream.from_byte_array(Data.fit_file_short)
+        decoder = Decoder(stream)
+        decoder.read()
+
+        assert stream.position() == stream.get_length()
+        messages, errors = decoder.read()
+        assert len(errors) == 0 and len(messages) == 0
+
+    def test_compressed_timestamp_message_should_throw(self):
+        '''Tests that the decoder should throw an error when reading a message with a compressed timestamp'''
+        stream = Stream.from_byte_array(Data.fit_file_short_compressed_timestamp)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read()
+
+        assert len(errors) == 1
+        assert "Compressed timestamp messages are not currently supported" in str(errors[0])
+
     def test_read_incorrect_field_def_size(self):
         '''Tests that the decoder doesn't break when reading a message with an incorrect field definition size.'''
         stream = Stream.from_byte_array(Data.fit_file_short_with_wrong_field_def_size)
@@ -103,6 +224,15 @@ class TestDecoderRead():
         assert len(errors) == 0
         assert "time_created" in messages["file_id_mesgs"][0]
 
+    def test_invalid_crc_should_fail(self):
+        '''Test decoder should fail when CRC is invalid'''
+        stream = Stream.from_byte_array(Data.fit_file_short_invalid_CRC)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read()
+
+        assert len(errors) == 1
+        assert len(messages['file_id_mesgs']) == 1
+
 
     @pytest.mark.parametrize(
         "data,expected_output",
@@ -227,7 +357,6 @@ class TestDecoderRead():
         '''Tests the validity of reading developer data from a fit file'''
         stream = Stream.from_file('tests/fits/ActivityDevFields.fit')
         decoder = Decoder(stream)
-        decoder.read()
         messages, errors = decoder.read()
 
         assert len(errors) == 0
@@ -240,11 +369,53 @@ class TestDecoderRead():
 
         stream = Stream.from_byte_array(Data.fit_file_dev_data_missing_field_description)
         decoder = Decoder(stream)
-        decoder.read()
         messages, errors = decoder.read()
 
         assert len(errors) == 0 and len(messages['activity_mesgs']) == 1
 
+    @pytest.mark.parametrize(
+        "option_status",
+        [
+            (True),
+            (False),
+            (None)
+        ], ids=["Set to True", "Set to False", "Default should have CRC calculations enabled"]
+    )
+    def test_enable_crc_options(self, mocker, option_status):
+        '''Tests enabling and disabling CRC calculation when decoding a FIT file.'''
+        spy_add_bytes = mocker.spy(CrcCalculator, "add_bytes")
+        spy_get_crc = mocker.spy(CrcCalculator, "get_crc")
+
+        stream = Stream.from_byte_array(Data.fit_file_short)
+        decoder = Decoder(stream)
+
+        if option_status is not None:
+            messages, errors = decoder.read(enable_crc_check=option_status)
+        else:
+            messages, errors = decoder.read()
+
+        assert len(errors) == 0
+
+        assert spy_add_bytes.call_count == 0 if option_status is False else spy_add_bytes.call_count > 0
+        assert spy_get_crc.call_count == 0 if option_status is False else spy_get_crc.call_count > 0
+
+    @pytest.mark.parametrize(
+        "option_status, data, expected_error_status",
+        [
+            (True, Data.fit_file_short_new, False),
+            (True, Data.fit_file_short_new_invalid_crc, True),
+            (False, Data.fit_file_short_new, False),
+            (False, Data.fit_file_short_new_invalid_crc, False),
+        ], ids=["With CRC | Valid File", "With CRC | Invalid File", "Without CRC | Valid File", "Without CRC | Invalid File"]
+    )
+    def test_enable_crc_options_errors_returned(self, option_status, data, expected_error_status):
+        '''Tests if errors are returned when decoding a file when CRC calculations are enabled or disabled.'''
+        stream = Stream.from_byte_array(data)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read(enable_crc_check=option_status)
+
+        assert len(errors) == 0 if expected_error_status is False else len(errors) > 0
+
     @pytest.mark.parametrize(
         "option_status",
         [
@@ -310,6 +481,23 @@ class TestDecoderRead():
         for mesg, distance in zip(duration_distance_workout_step_mesgs, distances):
             assert mesg['duration_distance'] == distance
 
+    def test_messages_with_no_fields(self):
+        '''Tests reading messages with no fields assigned in their message definition'''
+        stream = Stream.from_byte_array(Data.fit_file_messages_with_no_fields)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read()
+        assert len(errors) == 0
+
+        serial_number = 3452116910
+
+        assert len(messages['pad_mesgs']) == 1
+        assert len(messages['file_id_mesgs']) == 2
+
+        assert messages['pad_mesgs'][0] == {}
+
+        assert messages['file_id_mesgs'][0]["serial_number"] == serial_number
+        assert messages['file_id_mesgs'][1]["serial_number"] == serial_number
+
 class TestComponentExpansion:
     def test_sub_field_and_component_expansion(self):
         stream = Stream.from_file('tests/fits/WithGearChangeData.fit')
@@ -385,14 +573,19 @@ class TestComponentExpansion:
         stream = Stream.from_byte_array(Data.fit_file_monitoring)
         decoder = Decoder(stream)
         messages, errors = decoder.read()
-
         assert len(errors) == 0
         assert len(messages['monitoring_mesgs']) == 4
+        assert messages['monitoring_mesgs'][0]['activity_type'] == "running" and messages['monitoring_mesgs'][0]['intensity'] == 3
+        assert messages['monitoring_mesgs'][0]['cycles'] == 10
+
+        assert messages['monitoring_mesgs'][1]['activity_type'] == "walking" and messages['monitoring_mesgs'][1]['intensity'] == 0
+        assert messages['monitoring_mesgs'][1]['cycles'] == 30
+
+        assert messages['monitoring_mesgs'][2]['activity_type'] == 30 and messages['monitoring_mesgs'][2]['intensity'] == 0
+        assert messages['monitoring_mesgs'][2]['cycles'] == 15
 
-        assert messages['monitoring_mesgs'][0]['activity_type'] == 8 and messages['monitoring_mesgs'][0]['intensity'] == 3
-        assert messages['monitoring_mesgs'][1]['activity_type'] == 0 and messages['monitoring_mesgs'][1]['intensity'] == 0
-        assert messages['monitoring_mesgs'][2]['activity_type'] == 30 and messages['monitoring_mesgs'][2]['intensity'] == 6
         assert 'activity_type' not in messages['monitoring_mesgs'][3] and 'intensity' not in messages['monitoring_mesgs'][3]
+        assert messages['monitoring_mesgs'][3]['cycles'] == 15
 
 class TestMergeHeartrates:
     '''Set of tests which verify the functionality of merging heartrates to records when decoding.'''
@@ -436,6 +629,76 @@ class TestMergeHeartrates:
         messages, errors = decoder.read(expand_components=False)
         assert len(errors) == 1
 
+class TestAccumulatedFields:
+    def test_expanded_components_expand_with_fields_that_accumulate(self):
+        '''Tests that expanding components which are set to accumulate, accumulate properly.'''
+        stream = Stream.from_byte_array(Data.fit_file_accumulated_components)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read()
+        assert len(errors) == 0
+
+        assert messages['record_mesgs'][0]['cycles'] == 254
+        assert messages['record_mesgs'][0]['total_cycles'] == 254
+
+        assert messages['record_mesgs'][1]['cycles'] == 0
+        assert messages['record_mesgs'][1]['total_cycles'] == 256
+
+        assert messages['record_mesgs'][2]['cycles'] == 1
+        assert messages['record_mesgs'][2]['total_cycles'] == 257
+
+    def test_expanded_components_which_accumulate_and_have_initial_value_scale_and_accumulate(self):
+        '''Tests that when an accumulated, expanded component field that is given an initial value is scaled accordingly in accumulation.'''
+        stream = Stream.from_byte_array(Data.fit_file_compressed_speed_distance_with_initial_distance)
+        decoder = Decoder(stream)
+        messages, errors = decoder.read()
+        assert len(errors) == 0
+
+        # The first distance field is not expanded from a compressedSpeedDistance field
+        assert messages['record_mesgs'][0]['distance'] == 2
+        assert messages['record_mesgs'][1]['distance'] == 264
+        assert messages['record_mesgs'][2]['distance'] == 276
+        
+class TestDecoderExceptions:
+    '''Set of tests which verifies behavior of the decoder when various exceptions are raised'''
+    @pytest.mark.parametrize(
+        "exception",
+        [
+            KeyboardInterrupt,
+            SystemExit,
+        ], ids=["KeyboardInterrupt", "SystemExit"]
+    )
+    def test_keyboard_interrupt_and_system_exit_exceptions_are_rethrown(self, mocker, exception):
+        '''Tests to ensure that the decoder rethrows KeyboardInterrupt and SystemExit exceptions'''
+        stream = Stream.from_byte_array(Data.fit_file_short)
+        decoder = Decoder(stream)
+        
+        mocked_is_fit = mocker.patch('garmin_fit_sdk.Decoder.is_fit')
+        mocked_is_fit.side_effect = exception
+
+        with pytest.raises(exception): 
+            decoder.read()
+
+    @pytest.mark.parametrize(
+        "exception",
+        [
+            Exception,
+            RuntimeError,
+            BufferError,
+            LookupError,
+            IndexError
+        ], ids=["Generic Exception", "RuntimeError", "BufferError", "LookupError", "IndexError"]
+    )
+    def test_other_exceptions_are_not_rethrown(self, mocker, exception):
+        '''Tests to ensure that the decoder does not rethrow other exceptions'''
+        stream = Stream.from_byte_array(Data.fit_file_short)
+        decoder = Decoder(stream)
+        
+        mocked_is_fit = mocker.patch('garmin_fit_sdk.Decoder.is_fit')
+        mocked_is_fit.side_effect = exception
+
+        messages, errors = decoder.read()
+
+        assert len(errors) == 1
 
 def test_mesg_listener():
     '''Tests that a message listener passed to the decoder is correctly called.'''
